package vegr

import (
	"fmt"
	"reflect"
	"time"

	"git.sr.ht/~nelsam/hel/sel"
	"git.sr.ht/~nelsam/hel/vegr/ret"
)

type boolSet []bool

func newBoolSet(length int) boolSet {
	return make(boolSet, length)
}

func (s boolSet) all() bool {
	for _, v := range s {
		if !v {
			return false
		}
	}
	return true
}

type T interface {
	Helper()
	Failed() bool
	Logf(string, ...any)
	Fatalf(string, ...any)
	Cleanup(func())
}

func PopulateReturns(t T, name string, timeout time.Duration, mock any, addrs ...any) {
	t.Helper()

	doneCh := make(chan struct{})
	t.Cleanup(func() { <-doneCh })
	defer func() {
		defer close(doneCh)
		if r := recover(); r != nil {
			if t.Failed() {
				return
			}
			panic(r)
		}
	}()
	cases, err := sel.Cases(reflect.SelectRecv, mock)
	if err != nil {
		panic(fmt.Errorf("hel: PopulateReturns was called with incorrect mock type (%T): %w", mock, err))
	}
	expected := len(addrs)
	blockIdx := ret.BlockerFieldIdx(mock)
	if blockIdx >= 0 {
		expected++
	}
	panicIdx := ret.PanickerFieldIdx(mock)
	if panicIdx >= 0 {
		expected++
	}
	if len(cases) != expected {
		panic(fmt.Errorf("hel: PopulateReturns was called with %d channels but only %d addresses", len(cases), len(addrs)))
	}
	var vals []reflect.Value
	for _, a := range addrs {
		v := reflect.ValueOf(a)
		if v.Kind() != reflect.Ptr {
			panic(fmt.Errorf("hel: PopulateReturns was called with non-pointer type (%T)", a))
		}
		if v.IsNil() {
			panic(fmt.Errorf("hel: PopulateReturns was called with nil pointer of type (%T)", a))
		}
		vals = append(vals, v.Elem())
	}
	done := newBoolSet(len(cases))
	deadline := time.NewTimer(timeout)
	defer deadline.Stop()

	timeoutIdx := len(cases)
	cases = append(cases, reflect.SelectCase{
		Dir:  reflect.SelectRecv,
		Chan: reflect.ValueOf(deadline.C),
	})
	if t.Failed() {
		return
	}
	for !done.all() {
		chosen, recv, ok := reflect.Select(cases)
		if !ok {
			panic(fmt.Errorf("hel: PopulateReturns called on closed mock (type %T)", mock))
		}
		if t.Failed() {
			return
		}
		if chosen == timeoutIdx {
			defer func() {
				if r := recover(); r != nil {
					panic(fmt.Errorf("hel: panic calling t.Fatalf from mock method %v: %v", name, r))
				}
			}()
			warnPanicChanges(t, panicIdx, done)
			t.Fatalf("hel: mock method %v timed out after %v waiting for return on type (%T)", name, timeout, mock)
			return
		}
		done[chosen] = true
		cases[chosen] = reflect.SelectCase{
			Dir:  reflect.SelectRecv,
			Chan: reflect.ValueOf(chan struct{}(nil)), // blocks indefinitely
		}
		if chosen == blockIdx {
			continue
		}
		if chosen == panicIdx {
			val := recv.Interface()
			if val != nil {
				defer panic(val)
			}
			continue
		}
		valIdx := chosen
		if blockIdx >= 0 && chosen > blockIdx {
			valIdx--
		}
		if panicIdx >= 0 && chosen > panicIdx {
			valIdx--
		}
		vals[valIdx].Set(recv)
	}
	if blockIdx >= 0 {
		reflect.ValueOf(mock).Field(blockIdx).Send(reflect.ValueOf(struct{}{}))
	}
}

func warnPanicChanges(t T, panicIdx int, changes boolSet) {
	t.Helper()

	failed := -1
	for i, c := range changes {
		if c {
			continue
		}
		if failed != -1 {
			return
		}
		failed = i
	}
	if failed != panicIdx {
		return
	}
	t.Logf("hel: BREAKING CHANGE: as of v0.6, a value must also be sent on the MockPanicker field, even if a method has no return values [hint: use pers.Return, even with no return values, to ensure better forward compatibility]")
}
